// --*- C++ -*------x---------------------------------------------------------
// $Id: clustervectors2.cc,v 1.2 2007/08/14 18:27:50 bindewae Exp $
//
// Program:         - 
//
// Author:          Eckart Bindewald
//
// Project name:    -
//
// Date:            $Date: 2007/08/14 18:27:50 $
//
// Description:     
// 
// -----------------x-------------------x-------------------x-----------------

#define __STL_NO_DRAND48

#include <iostream>
#include <fstream>
#include <string>
#include <Vec.h>
#include <debug.h>
#include <GetArg.h>
#include <FileName.h>
#include <vectornumerics.h>
#include <clusterAlgorithms.h>
#include <Random.h>
#include <algorithm>

void
helpOutput(ostream& os)
{
  os << "clustervectors2 (-i filename|--matrix filename) -d cutoff " << endl
     << "--mask id1 id2 ...   : input column ids" << endl
     << "--max value  : maximum size of cluster" << endl
     << "--min value  : minimum size of cluster" << endl
     << "--root filename : root directory" << endl
     << "--unique 0|1    : unique data only" << endl
     << "--verbose level : verbose level"  << endl;  
}

/** output of command line parameter with which the program was called. */
void
parameterOutput(ostream& os, int argc, char** argv)
{
  for (int i = 0; i < argc; i++)
    {
      os << argv[i] << " ";
    }
  os << endl;
}

Vec<Vec<double> >
generateDistanceMatrix(const Vec<Vec<double> >& dataVecs) 
{
  Vec<double> distanceRow(dataVecs.size(), 0.0);
  Vec<Vec<double> > distances(dataVecs.size(), distanceRow);
  for (unsigned int i = 1; i < dataVecs.size(); ++i) {
    for (unsigned int j = 0; j < i; ++j) {
      distances[i][j] = euclidianDistance(dataVecs[i], dataVecs[j]);
      distances[j][i] = distances[i][j];
    }
  }
  return distances;
}

Vec<Vec<unsigned int> >
clusterVectors(const Vec<Vec<double> >& dataVecs,
	       double cutoff,
	       int clusterAlgorithm)
{
  Vec<Vec<unsigned int> > result;
  switch (clusterAlgorithm) {
  case 0:  // trivial clustering
    result = Vec<Vec<unsigned int> >(dataVecs.size());
    for (unsigned int i = 0; i < dataVecs.size(); ++i) {
      result[i] = Vec<unsigned int>(1, i);
    }
    break;
  case 1:
    result = simpleRepresentativeLinkage(dataVecs, cutoff); 
    break;
  default:
    ERROR("Unkonwn cluster algorithm!");
  }
  return result;
}

/** reads input file, returns table with each row of string corresponding to 
    one line of the input stream
*/
Vec<Vec<string> >
readStringTable(istream& is)
{
  Vec<Vec<string> > result;
  while (is) {
    string line = getLine(is);
    vector<string> words = getTokens(line);
    if (words.size() > 0) { // ignore lines with only whitespace
      result.push_back(words);
    }
  }
  return result;
}

Vec<Vec<string> >
readStringData(istream& is, int dataFormat)
{
  Vec<Vec<string> > result;
  switch (dataFormat) {
  case 0: is >> result;
    break;
  case 1:
    result = readStringTable(is);
    break;
  default:
    ERROR("Unknown input data format!");
  }
  return result;
}

Vec<double>
convertStringsToDoubleData(const Vec<string>& dataWords,
			   const Vec<unsigned int>& mask)
{
  PRECOND(mask.size() > 0);
  Vec<double> result(mask.size());
  for (unsigned int i = 0; i < mask.size(); ++i) {
    ERROR_IF(mask[i] >= dataWords.size(),
	     "Insuficient number of columns in data row!");
    result[i] = stod(dataWords[mask[i]]);
  }
  return result;
}

Vec<Vec<double> >
convertStringsToDoubleData(const Vec<Vec<string> >& dataWords,
			   Vec<unsigned int> mask)
{
  ERROR_IF((dataWords.size() == 0) || (dataWords[0].size() == 0),
	   "Could not find suitable first data line.");
  unsigned int numCol = dataWords[0].size();
  unsigned int goodCount = 1;
  for (unsigned int i = 1; i < dataWords.size(); ++i) {
    if (dataWords[i].size() == numCol) {
      ++goodCount;
    }
  }
  Vec<Vec<double> > result(goodCount);
  if (mask.size() == 0) {
    mask = generateStair(numCol);
  }
  unsigned int pc = 0;
  for (unsigned int i = 0; i < dataWords.size(); ++i) {
    if (dataWords[i].size() != numCol) {
      continue;
    }
    result[pc++] = convertStringsToDoubleData(dataWords[i], mask);
  }
  return result;
}

int
convertStringsToClassData(const Vec<string>& dataWords,
			  unsigned int dataCol,
			  Vec<string>& wordsSoFar)
{
  ERROR_IF(dataCol >= dataWords.size(), "Undefined class column data!");

  const string& newWord = dataWords[dataCol];
  for (unsigned int i = 0; i < wordsSoFar.size(); ++i) {
    if (wordsSoFar[i].compare(newWord) == 0) {
      return static_cast<int>(i);
    }
  }
  wordsSoFar.push_back(newWord);
  return static_cast<int>(wordsSoFar.size()-1);
}

Vec<int>
convertStringsToClassData(const Vec<Vec<string> >& dataWords,
			  unsigned int classCol)
{
  ERROR_IF((dataWords.size() == 0) || (dataWords[0].size() == 0),
	   "Could not find suitable first data line.");
  if (classCol == 0) {
    classCol == dataWords[0].size()-1;
  }
  unsigned int numCol = dataWords[0].size();
  unsigned int goodCount = 1;
  for (unsigned int i = 1; i < dataWords.size(); ++i) {
    if (dataWords[i].size() == numCol) {
      ++goodCount;
    }
  }
  Vec<int> result(goodCount);
  Vec<string> wordsSoFar;
  unsigned int pc = 0;
  for (unsigned int i = 0; i < dataWords.size(); ++i) {
    if (dataWords[i].size() != numCol) {
      continue;
    }
    result[pc++] = convertStringsToClassData(dataWords[i], classCol, wordsSoFar);
  }
  return result;
}

void
writeStringTable(ostream& os, 
		 const Vec<Vec<string> >& dataWords,
		 const Vec<Vec<unsigned int> >& clusters,
		 int clusterSizeMax,
		 int clusterSizeMin)
{
  Random& rnd = Random::getInstance();
  for (unsigned int i = 0; i < clusters.size(); ++i) {
    if (static_cast<int>(clusters.size()) < clusterSizeMin) {
      continue;
    }
    unsigned int maxOut = clusters[i].size();
    if (clusterSizeMax > 0) {
      maxOut = minimum(maxOut, static_cast<unsigned int>(clusterSizeMax));
    }
    Vec<unsigned int> rids = generateStair(maxOut);
    random_shuffle(rids.begin(), rids.end(), rnd);
    for (unsigned int jj = 0; jj < maxOut; ++jj) {
      unsigned int j = rids[jj]; // random index
      unsigned int id = clusters[i][j];
      for (unsigned int k = 0; k < dataWords[id].size(); ++k) {
	os << dataWords[id][k] << " ";
      }
      os << endl;
    }
  }
}


void
writeStringTable(ostream& os, 
		 const Vec<Vec<string> >& dataWords,
		 const Vec<unsigned int>& clusters)
{
  for (unsigned int i = 0; i < clusters.size(); ++i) {
    unsigned int id = clusters[i];
    for (unsigned int k = 0; k < dataWords[id].size(); ++k) {
      os << dataWords[id][k] << " ";
    }
    os << endl;
  }
}


void
writeData(ostream& os, 
	  const Vec<Vec<string> >& dataWords,
	  const Vec<Vec<unsigned int> >& clusters,
	  int outputFormat, 
	  int clusterSizeMax,
	  int clusterSizeMin)
{
  switch (outputFormat) {
  case 1: // same as input format "1":
    writeStringTable(os, dataWords, clusters, clusterSizeMax,
		     clusterSizeMin);
    break;
  default:
    ERROR("Unknown output format!");
  }
}

void
writeData(ostream& os, 
	  const Vec<Vec<string> >& dataWords,
	  const Vec<unsigned int>& ids,
	  int outputFormat)
{
  switch (outputFormat) {
  case 1: // same as input format "1":
    writeStringTable(os, dataWords, ids);
    break;
  default:
    ERROR("Unknown output format!");
  }
}



int
main(int argc, char ** argv)
{

  bool helpMode;
  int argcFile = 0;
  int clusterAlgorithm = 1; // 0: do nothing, 1: simpleRepresentative
  int clusterSizeMax = -1; // if > 0: restrict maximum size of output cluster
  int clusterSizeMin = 1; // minimum size of output cluster
  int inputFormat = 1; // 0: native format, 1: simple list
  int kk = 0; // k for k nearest neighbors
  int knn = -1; // run knn algorithm on which data point
  int maskMode = 1; // 1: all entries, 2: all but last columns
  int outputFormat = 1; // 0: native format, 1: simple list
  int uniqueMode = 0;
  char ** argvFile = 0;
  unsigned int classCol = 99999; // large number, user has to specify
  unsigned int verboseLevel = 0;
  double cutoff = 0.1;
  string commandFileName;
  string inputFileName;
  string logFileName; //  = "mainprogramtemplate.log";
  string rootDir = ".";
  Vec<unsigned int> mask;
  getArg("-help", helpMode, argc, argv);

  if ((argc < 2) || helpMode)  {
    helpOutput(cout);
    exit(0);
  }

  getArg("-root", rootDir, argc, argv, rootDir);
  addSlash(rootDir);

  getArg("-commands", commandFileName, argc, argv, commandFileName);
  addPathIfRelative(commandFileName, rootDir);

  if (commandFileName.size() > 0) {
    ifstream commandFile(commandFileName.c_str());
    if (!commandFile) {
      if (isPresent("-commands", argc, argv)) {
	ERROR_IF(!commandFile, "Error opening command file.");
      }
      else {
	cerr << "Warning: Could not find command file: " + commandFileName 
	     << endl;
      }
    }
    else {
      argvFile = streamToCommands(commandFile, argcFile, 
				  string("mainprogramtemplate"));
    }
    commandFile.close();
  }

  getArg("a", clusterAlgorithm, argcFile, argvFile, clusterAlgorithm);
  getArg("a", clusterAlgorithm, argc, argv, clusterAlgorithm);
  getArg("-class-col", classCol, argcFile, argvFile, classCol); // defines used class columns
  getArg("-class-col", classCol, argc, argv, classCol);
  --classCol; // internal counting starts at zero, external at one
  getArg("d", cutoff, argcFile, argvFile, cutoff);
  getArg("d", cutoff, argc, argv, cutoff);
  getArg("i", inputFileName, argc, argv, inputFileName);
  getArg("k", kk, argcFile, argvFile, kk);
  getArg("k", kk, argc, argv, kk);
  getArg("-knn", knn, argcFile, argvFile, knn);
  getArg("-knn", knn, argc, argv, knn);
  --knn;
  getArg("-log", logFileName, argc, argv, logFileName);
  getArg("-log", logFileName, argcFile, argvFile, logFileName);
  addPathIfRelative(logFileName, rootDir);
  getArg("-mask", mask, argcFile, argvFile); // defines used data columns
  getArg("-mask", mask, argc, argv);
  convert2InternalCounting(mask); // internal counting starts at zero
  getArg("-mask-mode", maskMode, argcFile, argvFile, maskMode); // defines used data columns
  getArg("-mask-mode", maskMode, argc, argv, maskMode);
  getArg("-max", clusterSizeMax, argcFile, argvFile, clusterSizeMax);
  getArg("-max", clusterSizeMax, argc, argv, clusterSizeMax);
  getArg("-min", clusterSizeMin, argcFile, argvFile, clusterSizeMin);
  getArg("-min", clusterSizeMin, argcFile, argvFile, clusterSizeMin);
  getArg("-unique", uniqueMode, argcFile, argvFile, uniqueMode);
  getArg("-unique", uniqueMode, argc, argv, uniqueMode);
  getArg("-verbose", verboseLevel, argcFile, argvFile, verboseLevel);
  getArg("-verbose", verboseLevel, argc, argv, verboseLevel);

  if (logFileName.size() > 0) {
    ofstream logFile(logFileName.c_str(), ios::app);
    parameterOutput(logFile, argc, argv);
    if (argcFile > 1) {
      logFile << "Parameters from command file: ";
      parameterOutput(logFile, argcFile, argvFile);
    }
    logFile.close();
  }
  
  /***************** MAIN PROGRAM *****************************/

  Vec<Vec<string> > dataWords;
  if (inputFileName.size() > 0) {
    ifstream inputFile(inputFileName.c_str());
    ERROR_IF(!inputFile, "Error opening input file!");
    dataWords = readStringData(inputFile, inputFormat);
    inputFile.close();
  }
  else { // read from standard input if no filename specified
    dataWords = readStringData(cin, inputFormat);
  }
  if (verboseLevel > 0) {
    cout << "# input data read: " << dataWords.size() << " lines!" << endl;
  }
  // if delete identical mode:
  if (uniqueMode) {
    cout << "Uniquifying data!" << endl;
    dataWords = uniqueSet(dataWords);
  }
  if (verboseLevel > 0) {
    cout << "# String data with " << dataWords.size() << " lines read."
	 << endl;
  }
  if (mask.size() == 0) {
    if (maskMode == 1) {
      mask = generateStair(dataWords[0].size()); // all but last column
    }
    else if (maskMode == 2) {
      mask = generateStair(dataWords[0].size()-1); // all but last column
    }
  }
  Vec<Vec<double> > dataVecs = convertStringsToDoubleData(dataWords, mask);
  if (verboseLevel > 0) {
    cout << "# Numeric data with " << dataVecs.size() << " lines generated."
	 << " Using mask: " << mask;
  }
  if (verboseLevel > 1) {
    cout << "# The input data is: " << dataVecs << endl;
  }

  if (knn >= 0) {
    Vec<int> dataClasses = convertStringsToClassData(dataWords, classCol);
    Vec<unsigned int> knnResult = kNearestNeighbors(dataVecs[knn],
						    dataVecs, kk);
    if (verboseLevel > 1) {
      cout << "Result off knn algorithm on datavec " << knn << " "
	   << dataVecs[knn] << " : " << knnResult;
    }
    for (unsigned int i = 0; i < knnResult.size(); ++i) {
      cout << dataVecs[knnResult[i]] << " " << euclidianDistance(dataVecs[knn],
		 dataVecs[knnResult[i]]) << " " << dataClasses[knnResult[i]] << endl;
    }
    exit(0);
  }

  if (kk > 0) {
    if (verboseLevel > 0) {
      cout << "#  Starting knn thinning with k = " << kk << endl;
    }
    Vec<int> dataClasses = convertStringsToClassData(dataWords, classCol);
    Vec<unsigned int> thinIds = knnThin(dataVecs, dataClasses, kk);
    if (verboseLevel > 0) {
      cout << "# Number of resulting data rows: " << thinIds.size() << endl;
    }
    dataVecs = getSubset(dataVecs, thinIds);
    dataClasses = getSubset(dataClasses, thinIds);
    writeData(cout, dataWords, thinIds, outputFormat);
    exit(0);
  }

  if (verboseLevel > 0) {
    cout << "# Starting clustering with algorithm " << clusterAlgorithm 
	 << " !" << endl;
  }
  Vec<Vec<unsigned int> > clusters = clusterVectors(dataVecs,
		    cutoff, clusterAlgorithm);

  if (verboseLevel > 0) {
    cout << "# Number of found clusters: " << clusters.size() << endl;
  }

  // find largest cluster
  unsigned int bestIds = 0;
  unsigned  int bestSize = clusters[0].size();
  for (unsigned int i = 1; i < clusters.size(); ++i) {
    if (clusters[i].size() > bestSize) {
      bestSize = clusters[i].size();
      bestIds = i;
    }
  }

  // find lowest line index in best cluster:
  // find largest cluster
  unsigned int bestIds2 = 0;
  unsigned  int lowest = clusters[bestIds][0];
  for (unsigned int i = 1; i < clusters[bestIds].size(); ++i) {
    if (clusters[bestIds][i] < lowest) {
      lowest = clusters[bestIds][i];
      bestIds2 = i;
    }
  }
  if (verboseLevel > 0) {
    cout << "# Found clusters: " << clusters << endl;  
    cout << "# Representative of largest cluster: " << lowest << " " 
	 << dataVecs[lowest] << endl;
  }
  writeData(cout, dataWords, clusters, outputFormat, 
	    clusterSizeMax, clusterSizeMin);

  return 0;
}
